#
# Tu Epm - Tcl Usefulness Edit Program Menu
#
# Keeps track of two menus that can be used by a program:
#   An epm menu with items for choosing files etc., and edit items.
#   An epm edit menu with only edit items.
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#
# Fall 2022
# Version 0.2
#

package require Tcl 8.5

namespace eval tu { namespace eval epm {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::epm
		title	{Tu Epm}
		descr	{Edit Program Menu}
		author	{Stuart Cassoff}
		when	{Fall 2022}
		version	0.2
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


### Procedures copied from other places

namespace eval tu {
namespace eval tcl {
proc randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

# dict get with default
#
proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

namespace export randstr dgwd
}; # namespace eval tcl

namespace eval tk {
#
# Underline Ampersand
# Convert ampersand in text to -underline option
# From Gub
#
proc ua {text} {	
	set index 0
	set uli -1
	while {[set index [string first "&" $text $index]] != -1} {
		if {[string index $text $index+1] eq "&"} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			incr index
			continue
		}
		if {$index < [string length $text]-1} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			set uli $index
		}
		incr index
	}
	set res [list $text]
	if {$uli != -1} {
		lappend res -underline $uli
	}
	return $res
}

namespace export ua
}; # namespace eval tk
}; # namespace eval tu


### Util

proc loadFileDialog {args} {
	if {[package provide tu::file] ne {}} {
		tailcall ::tu::file::loadFileDialog {*}$args
	} else {
		tailcall tk_getOpenFile {*}$args

	}
}
proc chooseDirectoryDialog {args} {
	if {[package provide tu::file] ne {}} {
		tailcall ::tu::file::chooseDirectoryDialog {*}$args
	} else {
		tailcall tk_chooseDirectory {*}$args
	}

}
proc saveSelectionToFile {args} {
	if {[package provide tu::file] ne {}} {
		tailcall ::tu::file::saveSelectionToFile {*}$args
	} else {
	}
}


### Epm

variable M {} ;# The epm menu
variable Mmap [dict create] ;# Named items and their indexes
variable E {} ;# The epm edit menu
variable Emap [dict create] ;# Named items and their indexes
# The 'mu' for E
variable epmmu [dict create -win {} -mmap {} -defs {} -ftmap {} -inserter {} -getwrap {} -setwrap {}]
variable epmww ;# Current word wrap state


# Create an epm menu
proc epmCreateMenu {{w ""} args} {
	variable M
	variable Mmap
	set M [menu $w.epm -tearoff 0 -postcommand [namespace current]::epmAdjustMenu {*}$args]
	foreach item [list "Choose &File" "Choose &Directory" "Choose C&olour" "&Set To Default" Cu&t &Copy &Paste C&lear "Select &All"] {
		$M add command -command [namespace current]::epm[string map {{ } {}} [string map {"&" ""} $item]] -label {*}[ua $item]
	}
	$M add checkbutton -command [namespace current]::epmAdjustWordWrap -variable [namespace current]::epmww -onvalue word -offvalue none -label {*}[ua "&Word Wrap"]
	$M insert 4 separator
	set Mmap [dict create cut 5 copy 6 paste 7 clear 8 wrap 10 setdef 3]
	return $M
}

# Return the epm menu, create if doesn't exist
proc epmMenu {} {
	variable M
	if {$M eq {}} { epmCreateMenu }
	return $M
}

# Create an edit menu
proc epmCreateAnEditMenu {{w ""} args} {
	set e [menu $w.eepm_[randstr] -tearoff 0 {*}$args]
	foreach item [list Cu&t &Copy &Paste C&lear "Select &All" "Sa&ve As"] {
		$e add command -command [namespace current]::epm[string map {{ } {}} [string map {"&" ""} $item]] -label {*}[ua $item]
	}
	$e add checkbutton -command [namespace current]::epmAdjustWordWrap -variable [namespace current]::epmww -onvalue word -offvalue none -label {*}[ua "&Word Wrap"]
	set emap [dict create cut 0 copy 1 paste 2 clear 3 saveas 5 wrap 6]
	epmSetEpmAdjustMenu $e $emap
	return [list $e $emap]
}

# Setup epm adjustment for a menu
proc epmSetEpmAdjustMenu {m mmap} {
	$m configure -postcommand [list [namespace current]::epmAdjustMenu $m $mmap]
}

# Create the epm edit menu
proc epmCreateEditMenu {{w ""}} {
	variable E
	variable Emap
	lassign [epmCreateAnEditMenu $w] E Emap
	return $E
}

# Return the epm edit menu, create if doesn't exist
proc epmEditMenu {} {
	variable E
	if {$E eq {}} { epmCreateEditMenu }
	return $E
}

# Have the epm menu popup for a widget
proc epmBindEpmMenuTo {w args} {
	variable Mmap
	# Call epmMenu to ensure M's menu exists
	epmBindMenuTo $w [epmMenu] {*}$args -mmap $Mmap
}

# Have the epm edit menu popup for a widget
proc epmBindEditMenuTo {w args} {
	variable Emap
	# Call epmEditMenu to ensure E's menu exists
	epmBindMenuTo $w [epmEditMenu] {*}$args -mmap $Emap
}

# Make a 'mu'
proc epmMakeMu {m args} {
	set ns [namespace current]:
	set mu [dict merge $args [dict create -menu $m \
		-mmap     [dgwd [dict create]      $args -mmap]     \
		-defs     [dgwd [dict create]      $args -defs]     \
		-ftmap    [dgwd [dict create]      $args -ftmap]    \
		-inserter [dgwd $ns:epmInserter    $args -inserter] \
		-getwrap  [dgwd $ns:epmGetWordWrap $args -getwrap]  \
		-setwrap  [dgwd $ns:epmSetWordWrap $args -setwrap]  \
	]]
	return $mu
}

# Have a menu popup for a widget
proc epmBindMenuTo {w m args} {
	set mu [epmMakeMu $m {*}$args]
	set mod   [dgwd ""        $args -mod]
	set seq   [dgwd "Button-" $args -seq]
	set bn    [dgwd 3         $args -bn]
	bind $w <$mod$seq$bn> [list [namespace current]::epmMenuGo %X %Y %W $mu]
}

# Popup activator
proc epmMenuGo {x y w mu} {
	variable epmmu [dict merge $mu [dict create -win $w -x $x -y $y]]
	tk_popup [dict get $epmmu -menu] $x $y
}

# Return the 'mu'
proc epmMu {} {
	variable epmmu
	return $epmmu
}

# Current window for a popup
proc epmW {args} {
	variable epmmu
	if {[llength $args] == 1} {
		dict set epmmu -win [lindex $args 0]
	}
	return [dict get $epmmu -win]
}

# Get the word wrap state for a window. Used by the default 'mu'.
proc epmGetWordWrap {w} {
	return [$w cget -wrap]
}

# Get the word wrap state for a window. Used by the default 'mu'.
proc epmSetWordWrap {w wrap} {
	$w configure -wrap $wrap
}

# Adjust a window's word wrap state according to epm state
proc epmAdjustWordWrap {} {
	variable epmmu
	variable epmww
	if {[dict get $epmmu -setwrap] eq {}} { return }
	[dict get $epmmu -setwrap] [dict get $epmmu -win] $epmww
}

# Ajdust a menu according to a window's state
proc epmAdjustMenu {{m {}} {mmap {}}} {
	variable epmmu
	variable epmww
	set w [dict get $epmmu -win]
	if {$w eq "" || ![winfo exists $w]} {
		if {![catch {info level -1} res]} {
			puts "epmAdjustMenu !$res!"
		}
		return

	}

if {[dict get $epmmu -getwrap] eq {}} { return }
#if {[dict get $epmmu -setwrap] eq {}} { return }

	set defs [dict get $epmmu -defs]
	if {$m eq {}} {
		set m [dict get $epmmu -menu]
	}
	if {$mmap eq {}} {
		set mmap [dict get $epmmu -mmap]
	}
	set items [list]
	if {[dict size $mmap] > 0} { foreach what [list cut copy paste clear wrap setdef saveas] {
		if {![dict exists $mmap $what]} { continue }
		lappend items $what
	}}
	if {[llength $items] == 0} { return }
	set states [dict create]
	dict set states setdef [expr {[dict exists $defs $w] ? "normal" : "disabled"}]
	set wclass [winfo class $w]
	if {$wclass eq "Text"} {
		set epmww [[dict get $epmmu -getwrap] $w]
		if {$epmww eq "char"} { set epmww word }
		dict set states wrap normal
		set sel [expr {[llength [$w tag ranges sel]] > 0 ? "normal" : "disabled"}]
	} elseif {$wclass eq "TLabel"} {
		set epmww none
		dict set states wrap disabled
		set sel normal
	} elseif {$wclass eq "TButton"} {
		set epmww none
		dict set states wrap disabled
		set sel disabled
	} else {
		set epmww none
		dict set states wrap disabled
		set sel [expr {[$w selection present] ? "normal" : "disabled"}]
	}
	dict set states copy $sel
	dict set states saveas $sel
	if {[$w cget -state] eq "normal"} {
		foreach item [list cut clear] {
			dict set states $item $sel
		}
		dict set states paste normal
	} else {
		foreach item [list cut clear paste setdef saeveas] {
			dict set states $item disabled
		}
	}
	if {[dict get $states paste] eq "normal" && "paste" in $items} {
		if {[catch {selection get -displayof $w -selection CLIPBOARD} s]} { set s "" }
		dict set states paste [expr {[string length $s] > 0 ? "normal" : "disabled"}]
	}
	foreach item $items {
		$m entryconfigure [dict get $mmap $item] -state [dict get $states $item]
	}
}

proc epmInserter {where what} {
	epmSelectAll ; epmClear
	$where insert end $what
}

proc epmChooseFile {} {
	variable epmmu
	set w [dict get $epmmu -win]
	set ftmap [dict get $epmmu -ftmap]
	set filetypes [dgwd {} $ftmap $w]
	if {[lsearch -exact -index {0 1} $filetypes [list *]] == -1} {
		lappend filetypes [list "All Files" [list *]]
	}
	if {[set v [loadFileDialog -initialdir [file dirname [$w get]] -initialfile [$w get] -filetypes $filetypes]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}

proc epmChooseDirectory {} {
	variable epmmu
	set w [dict get $epmmu -win]
	if {[set v [chooseDirectoryDialog -initialdir [$w get]]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}

proc epmChooseColour {} {
	variable epmmu
	set w [dict get $epmmu -win]
	if {[winfo class $w] eq "Text"} {
		set v ""
	} else {
		set v [string trim [$w get]]
	}
	if {$v eq ""} { set v #000000 }
	if {[catch {winfo rgb . $v}]} { set v olive }
	if {[set v [tk_chooseColor -initialcolor $v]] eq ""} { return }
	[dict get $epmmu -inserter] $w $v
}

proc epmSaveAs {} {
	variable epmmu
	saveSelectionToFile [dict get $epmmu -win]
}

proc epmSetToDefault {} {
	variable epmmu
	set w [dict get $epmmu -win]
	set defs [dict get $epmmu -defs]
	if {![dict exists $defs $w]} { return }
	set v [dict get $defs $w]
	[dict get $epmmu -inserter] $w $v
}

foreach e [list Cut Copy Paste Clear SelectAll] {
	proc epm$e {} [string map [list @e@ $e] {
		variable epmmu
		set w [dict get $epmmu -win]
		event generate $w <<@e@>>
		focus $w
	}]
}; unset e


## Setup

namespace import tu::tcl::randstr tu::tcl::dgwd
namespace import tu::tk::ua

namespace export epmCreateMenu epmMenu
namespace export epmCreateEditMenu epmCreateAnEditMenu epmEditMenu
namespace export epmBindMenuTo epmBindEpmMenuTo epmBindEditMenuTo
namespace export epmMakeMu epmSetEpmAdjustMenu
namespace export epmMu epmW

package provide tu::epm 0.2

}}; # namespace tu::epm


# EOF
