#
# Tu File - Tcl Usefulness File Utils
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#
# Fall 2022
# Version 0.2
#

package require Tcl 8.5

namespace eval tu { namespace eval file {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::file
		title	{Tu File}
		descr	{File Utils}
		author	{Stuart Cassoff}
		when	{Fall 2022}
		version	0.2
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


### File

proc getFileFor {what {fn ""} args} {
	if {$fn ni {"" "?"}} { return $fn }
	if {[package provide Tk] eq ""} { return "" }
	return [${what}FileDialog {*}$args]
}

proc getLoadFile {{fn ""} args} {
	dict lappend args -filetypes [list "All Files" [list *]]
	return [getFileFor load $fn {*}$args]
}

proc getSaveFile {{fn ""} args} {
	dict lappend args -filetypes [list "All Files" [list *]]
	return [getFileFor save $fn {*}$args]
}

proc openFile {fn {mode r}} {
	if {$fn eq "-"} { return [expr {$mode eq "r" ? "stdin" : "stdout"}] }
	if {$fn in {stdin stdout stderr}} { return $fn }
	if {[catch {open $fn $mode} f]} { return "" }
	return $f
}

proc closeFile {f} {
	if {$f in {stdin stdout stderr}} { return }
	close $f
}

proc loadFile {{fn ""} args} {
	set fn [getLoadFile $fn {*}$args]
	if {$fn eq ""} { return [list "" "" 0] }
	if {[set f [openFile $fn]] eq ""} { return [list "" "" 0] }
	set data [read -nonewline $f]
	closeFile $f
	return [list $fn $data 1]
}

proc saveFile {fn data args} {
	set fn [getSaveFile $fn {*}$args]
	if {$fn eq ""} { return [list "" 0] }
	if {[set f [openFile $fn w]] eq ""} { return [list "" "" 0] }
	puts -nonewline $f $data
	closeFile $f
	return [list $fn 1]
}

proc loadListFile {{fn ""} args} {
	if {![lassign [loadFile $fn {*}$args] fn data]} { return [list $fn "" 0] }
	return [list $fn [processListFileBlob $data] 1]
}

proc processListFileBlob {blob} {
	set res [list]
	foreach l [split $blob \n] {
		set l [string trim $l]
		if {$l eq {} || [string index $l 0] eq {#}} { continue }
		if {[catch {set len [llength $l]}] || $len < 1} { continue }
		lappend res $l
	}
	return $res
}

proc filemode->text {mode} {
    foreach { mask pairs } {
	00400 { 00400 r }
	00200 { 00200 w }
	04100 { 04100 s 04000 S 00100 x }
	00040 { 00040 r }
	00020 { 00020 w }
	02010 { 02010 s 02000 S 00010 x }
	00004 { 00004 r }
	00002 { 00002 w }
	01001 { 01001 t 01000 T 00001 x }
    } {
	set value [expr $mode & $mask]
	set bit -
	foreach { x b } $pairs {
	    if { $value == $x } {
		set bit $b
	    }
	}
	append retval $bit
    }
    return $retval
}

proc loadFileDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_getOpenFile {*}$args
	} else {
		tailcall ttk::getOpenFile {*}$args
	}
}

proc saveFileDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_getSaveFile {*}$args
	} else {
		tailcall ttk::getSaveFile {*}$args
	}
}

proc chooseDirectoryDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_chooseDirectory {*}$args
	} else {
		tailcall ttk::chooseDirectory {*}$args
	}
}

proc loadFileIntoText {w {fn ""} args} {
	lassign [loadFile $fn {*}$args] fn data
	if {$fn eq ""} { return "" }
	$w delete 1.0 end
	$w insert end $data
	return $fn
}

proc saveSelectionToFile {w {fn ""} args} {
	if {[llength [set sel [$w tag ranges sel]]] != 2} { return }
	set blob [string trim [$w get {*}$sel]]
	if {$blob eq ""} { return }
	saveFile $fn $blob {*}$args
	return $fn
}


## Setup

namespace export openFile closeFile getLoadFile loadFile getSaveFile saveFile
namespace export loadListFile processListFileBlob
namespace export loadFileDialog saveFileDialog
namespace export loadFileIntoText saveSelectionToFile
namespace export filemode->text

package provide tu::file 0.2

}}; # namespace tu::file


# EOF
