#
# Tu Epm - Tcl Usefulness Tcl Utils
#
#
# Stuart Cassoff
#
# Fall 2022
# Version 0.1
#

package require Tcl 8.5

namespace eval tu { namespace eval tcl {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::tcl
		title	{Tu Tcl}
		descr	{Tcl Utils}
		author	{Stuart Cassoff}
		when	{Fall 2022}
		version	0.1
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


proc randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

# dict get with default
#
proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

# In My Namespace
#
proc imn {what} {
	return [string trimright [uplevel 1 {namespace current}] :]::$what
}

#
# String Cat
#
proc strcat {args} { join $args "" }

#
# [lmap] replacement if not available
#
if {[llength [info commands lmap]] == 0} { proc lmap args {
	# 8.5 doesn't have lmap
	# https://wiki.tcl-lang.org/page/lmap+forward+compatibility
	set body [lindex $args end]
	set args [lrange $args 0 end-1]
	set n 0
	set pairs [list]
	foreach {varnames listval} $args {
		set varlist [list]
		foreach varname $varnames {
			upvar 1 $varname var$n
			lappend varlist var$n
			incr n
		}
		lappend pairs $varlist $listval
	}
	set temp [list]
	foreach {*}$pairs {
		lappend temp [uplevel 1 $body]
	}
	set temp
}}; # lmap

proc getNextOrPrevMatch {aList aString {curIdx 0} {backwards no}} {
	set idxs [lsearch -all -nocase $aList $aString*]
	if {[llength $idxs] == 0} { return -1 }
	if {[llength $idxs] == 1 && [lindex $idxs 0] == $curIdx} { return -1 }
	set p -1
	foreach idx $idxs {
		if {$idx < $curIdx} { set p $idx }
		if {$idx > $curIdx} { break }
	}
	if {$backwards} {
		if {$p == -1} {
			set newIdx [lindex $idxs end]
		} else {
			set newIdx $p
		}	
	} else {
		if {$idx == $curIdx} {
			set newIdx [lindex $idxs 0]
		} else {
			set newIdx $idx
		}
	}
	return $newIdx
}

proc whereBe {where curIdx nitems {height {10}}} {
	set idx $curIdx
	switch -exact -- $where {
	home     {  set idx  0       }
	end      {  set idx  $nitems }
	next     { incr idx  1       }
	prev     { incr idx -1       }
	nextpage { incr idx  $height ; if {$idx > $nitems} { set idx $nitems } }
	prevpage { incr idx -$height ; if {$idx < 0      } { set idx 0       } }
	default  {  set idx  $where  }
	}
	if {$idx < 0 || $idx > $nitems || $idx == $curIdx} { set idx -1 }
	return $idx
}

proc laddAndSort {list value {sort {-dictionary}}} {
	set i -1
	set values [lsort {*}$sort -index 0 [lmap e [linsert $list end $value] { list $e [incr i] }]]
	return [list [lsearch -index 1 $values $i] [lmap e $values { lindex $e 0 }]]
}

proc makeRWAccessors {args} {
	set c ""
	foreach m $args { append c method " " $m " " args " {" \
		[string trimright [regsub -all -line {^ +} [string map [list @m@ $m] {
			if {[llength $args] == 1} { set @m@ [lindex $args 0] }
			return $@m@
		}] "\t"]] "\n}\n"
	}
	return [string trimright $c]
}
proc makeROAccessors {args} {
	set c ""
	foreach m $args { append c method " " $m " " "{}" " {" \
		[string trimright [regsub -all -line {^ +} [string map [list @m@ $m] {
			return $@m@
		}] "\t"]] "\n}\n"
	}
	return [string trimright $c]
}

#
# Remove blank lines and comment lines, add space-backslash to all lines but the last.
# If a line contains ";#" (spaces ok too) then chares from those to the end of the line will be removed
# Useful for making big lists in code without all the unpleasant backslashes.
#
proc backslashify {blob} { return [join [lmap l [split $blob \n] { if {[regexp {^[[:space:]]*(;?#.*)?$} $l]} { continue } else { lindex [regexp -inline -- {^(.+?)(;\s*#.*)?$} $l] 1 } }] " \\\n"] }

proc maxstrlen {args} { return [::tcl::mathfunc::max 0 {*}[lmap s $args { string length $s }]] }

proc oneOf {l} {
	if {[llength $l] < 2} { return [join $l] }
	return [strcat "one of " [join [lrange $l 0 end-1] ", "] [expr {[llength $l] > 2 ? "," : ""}] " or " [lindex $l end]]
}



## Setup

namespace export makeRWAccessors makeRWOccessors
namespace export randstr dgwd imn backslashify maxstrlen oneOf
namespace export getNextOrPrevMatch whereBe laddAndSort

package provide tu::tcl 0.1

}}; # namespace tu::tcl


# EOF
